! Fork based parallel algorithm implementation

Algorithm$Parallel$Fork = Class[{Algorithm$ParallelHelper},
  { SchedulingMargin = 2,
    MaxSplitChunkSize = 64,
    SharedSizeArrayOverhead = 24,
    SharedSizeHeaderSize = SharedSize[Rule[0, {{}}]] - SharedSize[{}],
    MinimumSharedSize = Max[SharedSize[Rule[0, 0]], SharedSize[Rule[0, {}]]],
    Shared$Dormant =  0,
    Reply$Result   =  1,
    Reply$Size     =  2,
    Command$Exec   = -1,
    Command$Eval   = -2
    }, {},

  (* Initialize *)
  InitializeAttributes[] := (Algorithm$ParallelHelper`InitializeAttributes[]);

  (* Public interface *)
  Map[f_, l_, opt:((_Rule|_RuleDelayed)...)] := If[Length[l] > 1,
    Map1[f, l, opt], Literal[Map][f, l]];

  MapThread[f_, l_, opt:((_Rule|_RuleDelayed)...)] := Map[f@@#&, Thread[l], opt];

  (* Internal implementation *)
  Map1[fun_, l_, opt:((_Rule|_RuleDelayed)...)] := Module[{Map = Literal[Map],
    $head = Head[l], $length = Length[l], $save$pgid = GetPGID[],
    $split = 1, $l,

    slot, pid, shm, level, size, buf,
    $slot, $shm, $pid, $status, $opt, $level, $need$eval,

    $npara = 0, $running = 0, $pgid = 0, shm$size, npara},

    If[$length == 0, Return[l]];

    With[{
      ncpu = NumberOfProcessors[],
      opt$shm$size = Literal[SharedSize]/.{opt},
      opt$npara    = Literal[Multiplicity]/.{opt} },
      shm$size     = Max[0,
	If[RealQ[opt$shm$size], opt$shm$size, DefaultSharedSize]];
      npara        = Max[1,
	If[RealQ[opt$npara], opt$npara, NPARA]];
      If[RealQ[ncpu], npara = Min[ncpu, npara]];
      ];

    If[npara < 2, Return[fun/@l]];

    $split = Which[
      $length < 4 * npara,	1,

      True, Module[{cost$min = Infinity, split = 1, $improve = 0.95},
	Scan[With[{cost = 0
	  + (# * Ceiling[Ceiling[$length / #] / npara])	(* process cost *)
	  + Ceiling[$length / #]			(* fork cost *)
	  + 1},
!!	  Print["COST: ", {$length, npara, #, cost, $improve*cost$min, split}];
	  If[cost < $improve * cost$min, cost$min = cost; split = #]]&,
	  Range[Min[MaxSplitChunkSize, Ceiling[$length / npara]], 1, -1]];
	Ceiling[$length / split]]];

!!    Print["DEBUG[SPLIT]: ", {$length, npara, $split}];

    If[$split > 1, (
      shm$size = (SharedSizeArrayOverhead + shm$size) * Ceiling[$length / $split];
      $l = Split[l, $split]),
      $split = 1];

    shm$size = Max[MinimumSharedSize, shm$size + SharedSizeHeaderSize];

!!    Print["DEBUG[SHM]: ", {Literal[SharedSize]/.{opt}, $split, shm$size}];

    slot[_]  := Null;	(* PID -> Task slot *)
    pid[_]   := Null;	(* Task slot -> PID *)
    shm[_]   := Null;	(* Task slot -> shared memory *)
    size[_]  := Null;	(* Task slot -> shared memory size *)
    level[_] := Null;	(* Task slot -> level specifier *)
    buf[___] := {};	(* Result buffer *)

    (* Allocate process group ID *)
    $pgid = GetPID[];
    If[$save$pgid <> $pgid, SetPGID[0, $pgid]];

!    Switch[$pid = Fork[],
!      -1,($pgid = GetPID[]; SetPGID[0, $pgid]),
!      0, (While[True, Sleep[10]]; Exit[0]),
!      _, ($pgid = $pid; SetPGID[$pid, $pgid])];

    ScanThread[With[{task$arg = #1, task$level = {#2}},
      Which[
	$npara < npara,	( (* Assign new task slot *)
	  $npara += 1;
	  $slot = $npara;
	  $shm  = OpenShared[shm$size];
	  size[$slot] = shm$size;
	  shm[$slot] = $shm),

	$running < $npara,	( (* Re-claim shm for dormant task slot *)
	  $slot = First[Select[Range[$npara], Not[RealQ[pid[#]]]&, 1]];
	  $shm  = shm[$slot]),

	True,			( (* Re-claim shm for running task slot *)
	  $slot = Null; While[$slot === Null,
	    $pid  = -1; While[$pid == -1, {$pid, $status, $opt} = Wait4[-$pgid]];
	    If[$pid == 0, If[$pgid <> GetPID[], $pgid == GetPID[], Break[]]];
	    $slot = slot[$pid]; slot[$pid] =.; pid[$slot] =.;];

	  If[$pid == 0,		(* Re-claim orphan co-process *)
	    With[{$orphan = Select[Range[$npara], And[
	      RealQ[pid[#]],
	      Not[Check[SetPGID[pid[#], $pgid]; True, False]],
	      True]&, 1]},

	      If[Length[$orphan] < 0,
		$slot = First[$orphan];
		$pid = pid[$slot]; slot[$pid] =.; pid[$slot] =.;]]];

	  $shm  = shm[$slot];

	  (* Re-claim results *)
	  $level = level[$slot];
	  With[{$result = Shared[$shm]},
	    buf[Null@@$level] = Switch[$result,
	      Rule[Reply$Result, _],	Second[$result],

	      Rule[Reply$Size, _Real],	(
		shm$size = Max[shm$size, Second[$result]];
		If[$split > 1,
		  fun/@Part[$l, Null@@$level],
		  {fun[Part[l, Null@@$level]]}]),

	      _,			(
		If[$split > 1,
		  fun/@Part[$l, Null@@$level],
		  {fun[Part[l, Null@@$level]]}])]];

	  $running -= 1)];

      (* Extend shared memory size if needed *)
      If[RealQ[$slot] && size[$slot] < shm$size,
	With[{new$shm = OpenShared[shm$size]}, If[RealQ[new$shm],
	  shm[$slot]  = new$shm;
	  size[$slot] = shm$size;
	  CloseShared[$shm];
	  $shm = new$shm]]];

      $level = task$level;
      $need$eval = True;
      If[RealQ[$shm],
	Shared[$shm] = Command$Exec->0;
	Switch[$pid = Fork[],
	  -1,(	(* Can't fork *)
	    Shared[$shm] = Shared$Dormant->0),

	  0, (
	    FFS$InterruptMask = 7;
	    SetPGID[0, $pgid];
	    With[{result = If[$split > 1, fun/@task$arg, {fun[task$arg]}]},
	      Check[Shared[$shm] = Reply$Result->result,
!!		Print["DEBUG[EXCEED]: ", SharedSize[Reply$Result->result]];
		Shared[$shm] = Reply$Size->SharedSize[Reply$Result->result]]];
	    Exit[0]),

	  _, (
	    slot[$pid]   = $slot;
	    pid[$slot]   = $pid;
	    level[$slot] = $level;
	    $need$eval = False;
	    $running += 1)]];

      If[$need$eval, buf[Null@@$level] = If[$split > 1,
	fun/@Part[$l, Null@@$level], {fun[Part[l, Null@@$level]]}]];
      ]&, If[$split > 1, {$l, Range[Length[$l]]}, {l, Range[$length]}]];

    (* Re-claim results *)
    While[$running > 0,
      $slot = Null; While[$slot === Null,
	$pid  = -1; While[$pid == -1, {$pid, $status, $opt} = Wait4[-$pgid]];
	If[$pid == 0, If[$pgid <> GetPID[], $pgid == GetPID[], Break[]]];
	$slot = slot[$pid]; slot[$pid] =.; pid[$slot] =.;];

      If[$pid == 0,		(* Re-claim orphan co-process *)
	With[{$orphan = Select[Range[$npara], And[
	  RealQ[pid[#]],
	  Not[Check[SetPGID[pid[#], $pgid]; True, False]],
	  True]&, 1]},

	  If[Length[$orphan] < 0,
	    $slot = First[$orphan];
	    $pid = pid[$slot]; slot[$pid] =.; pid[$slot] =.;]]];

      $shm  = shm[$slot];

      (* Re-claim results *)
      $level = level[$slot];
      With[{$result = Shared[$shm]},
	buf[Null@@$level] = Switch[$result,
	  Rule[Reply$Result, _],	Second[$result],

	  Rule[Reply$Size, _Real],	(
	    shm$size = Max[shm$size, Second[$result]];
	    If[$split > 1,
	      fun/@Part[$l, Null@@$level],
	      {fun[Part[l, Null@@$level]]}]),

	  _,				(
	    If[$split > 1,
	      fun/@Part[$l, Null@@$level],
	      {fun[Part[l, Null@@$level]]}])]];

      (* Release shm *)
      CloseShared[$shm];

      $running -= 1];

    (* Re-claim process groupd ID *)
!    If[$pgid > 0 && $pgid <> GetPID[], Kill[$pgid, SIGKILL]; Wait4[$pgid]];

    (* Restore process groupd ID *)
    If[$save$pgid <> $pgid, SetPGID[0, $save$pgid]];

    Switch[$head,
      List,	       Table[Null@@buf[$level], {$level, $length}],
      _,	$head@@Table[Null@@buf[$level], {$level, $length}]]];

  ];

(* Set Hold attribute for algorithm member functions *)
Algorithm$Parallel$Fork@InitializeAttributes[];

! End of File
