! CaMonitor frontend part for asynchronous EPICS CA interface
!
! CaMonitor class is originally developed by N. Akasaka since 1998(Packages/CaSad2c.n).
! CaMonitor class is overhauled for new EPICS$Ca backend at 2017.12 by A. Morita.
!
CaMonitor = Class[{},
  { rl,				(* Hash map chid -> {obj_CaMonitor, 0(Scalor)|n_PositiveInteger(Vector index)} *)
    lazyFlushAfter = Null,	(* After handle for Lazy FlushIO *)
    PendEventDelay = 5,		(* Minimum delay between DoPendEvent *)
    lastPendEvent = 0,		(* Last DoPendEvent timestamp *)
    afterPendEvent = Null,
    bcschk = False, il={},	(* for CheckConStat[] *)
    
    (* Constants for Programming *)
    EPICS$Epoch=FromDate[{1990,1,1,0,0,0,0}],				(* EPICS Epoch: 1990-01-01 00:00:00 UTC *)
    DBE$Value=1, DBE$Log=2, DBE$Alarm=4, DBE$Property=8,		(* DBE_* macro in caeventmask.h *)
    CS$NeverConn=0, CS$PrevConn=1, CS$Conn=2, CS$Closed=3,		(* enum channel_state in cadef.h *)
    SEV$None=0, SEV$Minor=1, SEV$Major=2, SEV$Invalid=3,		(* enum epicsAlarmSeverity in alarm.h *)
    DBF$String=0, DBF$Int=1, DBF$Short=2, DBF$Float=3, DBF$Enum=4,	(* DBF_* macro in db_access.h *)
    DBF$Char=5, DBF$Double=6,
    DBF$Native=-1,							(* IOC Native type for EPICS$CaAddEvent[] *)

    (* Constants for backword user code compatibility *)
    DB$NoAlarm=0,DB$MinorAlarm=1,DB$MajorAlarm=2,DB$InvalidAlarm=3,	(* Alias of SEV$* *)
    DBF$Real=6},							(* Alias of DBF$Double *)

  { (* Instance Options *)
    DefaultValueType = Automatic,	(* Default monitoring value type. Override by ValueType option *)
    DefaultCount = 0,			(* Default count for EPICS$CaAddEvent[]. Override by Count option *)
    AlarmOnly = False,			(* Monitoring alarm event only[Default: with Value] *)
    AutoStart = True,			(* Automatic start monitor at connection *)
    InitialValue = Undefined,		(* Initial value of Value field *)
    Flush = True,			(* CA request flush mode: *)
    (* True:	Immediate ca_flush_io *)
    (* Lazy:	Queuing ca_flush_io by AfterIdle *)
    (* False:	Don't ca_flush_io *)

    (* EPICS CA connection variables *)
    nrn, nact,				(* # of PV[-1: Scalar], # of active PVs *)
    chid = Null, evid,			(* CHID[Null: Unconfigured], EVID *)
    rn,  cs,  st,  sev,  ts,  v,	(* PV, CS$*, Status, SEV$*, TimeStamp, Value *)
    rnl, csl, stl, sevl, tsl, vl, pos,	(* PV, CS$*, Status, SEV$*, TimeStamp, Value, Position @ Last Event *)

    (* Decoded Options *)
    cscomm  = Null,	(* ConStatCommand *)
    valcomm = Null,	(* ValueCommand *)
    putcomm = Null,	(* PutCommand *)

    vtype, evflag, count, 	(* EPICS$CaAddEvent parameter cache *)

    (* Instance states *)
    waitflg = False},

  Instances[] := Union[SwitchCases[Extract[?rl, {1, , 2}, Hold], {Hold[{_, _}]}][[1, , 1, 1]]];
  Instances$[] := SwitchCases[Extract[?rl, {1, , 2}, Hold], {Hold[{_, _}]}][[1, , 1]];

  Constructor[pv_] := (
    (* FIXME:Instance variable CANNOT initialized by Undefined vis instance variable declaration block *)
    If[Not[BoundQ[InitialValue]], InitialValue = Undefined];

    evflag = BitwiseOr[DBE$Alarm, If[AlarmOnly, 0, DBE$Value]];

    CreateChannel[pv, Null@@Options];

    If[TkinterQ[], DoPendEvent[True]];	(* Restart DoPendEvent loop if needed *)

    If[bcschk, AppendTo[il, This]]);

  Destructor[] := (
    If[bcschk, il = DeleteCases[il, This]];
    ClearChannel[]);

  (* Flush transmission queue *)
  DoFlushIO$[] := (lazyFlushAfter = Null; EPICS$CaFlushIO[]);

  FlushIO[]           := EPICS$CaFlushIO[];			(* Flush immediate *)
  FlushIO[flush_Real] := EPICS$CaFlushIO[];			(* Flush immediate *)
  FlushIO[flush_]     := If[TkinterQ[], (
    FlushIO[flush$_] := And[lazyFlushAfter === Null,
      (lazyFlushAfter = AfterIdle[DoFlushIO$[]]; False)];	(* Queuing lazy flush *)
    FlushIO[flush]), EPICS$CaFlushIO[]];
  FlushIO[False] := False;					(* Don't flush *)
  LazyFlushIO[] := FlushIO[Flush];

  (* Create/Clear channel *)
  NewName[pv__] := CreateChannel[pv];

  ReconfigureOptions[opt:((_Rule|_RuleDelayed)...)] := Module[{cst, val, pcb, vt, cnt, ao, as, ini, syn},
    {cst, val, pcb, vt, cnt, ao, as, ini, syn} = SwitchCases[Options,
      { RuleDelayed[ConStatCommand, _], RuleDelayed[ValueCommand, _], RuleDelayed[PutCommand, _],
	Rule[Literal[ValueType]|Literal[DefaultValueType], _], Rule[Literal[Count]|Literal[DefaultCount], _],
	Rule[Literal[AlarmOnly], _Real], Rule[Literal[AutoStart], _Real],
	Rule[Literal[InitialValue], _], Rule[Literal[Flush], _]}];
    If[cst <=> {},  cscomm = Extract[cst, {1, 2}, Hold]];
    If[val <=> {}, valcomm = Extract[val, {1, 2}, Hold]];
    If[pcb <=> {}, putcomm = Extract[pcb, {1, 2}, Hold]];
    If[ vt <=> {}, DefaultValueType =  vt[[1, 2]]];
    If[cnt <=> {}, DefaultCount     = cnt[[1, 2]]];
    If[ ao <=> {}, evflag = BitwiseOr[DBE$Alarm, If[ao[[1, 2]], 0, DBE$Value]]];
    If[ as <=> {}, AutoStart    =  as[[1, 2]]];
    If[ini <=> {}, InitialValue = ini[[1, 2]]];
    If[syn <=> {}, Flush        = syn[[1, 2]]];
    ];

  CreateChannel[pv_List, opt:((_Rule|_RuleDelayed)...)] := (
    ClearChannel[];

    ReconfigureOptions[opt];

    rn  = Map[If[StringQ[#] && StringLength[#] > 0, #, Null]&, pv];
    nrn = Length[rn];

    cs  = Table[CS$NeverConn, {nrn}];
    st  = Table[Undefined, {nrn}];
    sev = Table[SEV$None, {nrn}];
    ts  = Table[0, {nrn}];
    v   = Table[InitialValue, {nrn}];

    Clear[rnl, csl, sevl, tsl, vl, pos];

    chid = MapThread[If[StringQ[#2], With[{id = Check[EPICS$CaOpen[#2], Null]},
      If[RealQ[id], rl[id] = {This, #1}]; id], Null]&, {Range[nrn], rn}];

    nact = Plus@@(RealQ/@chid);

    (* Expand option for EPICS$CaAddEvent[] *)
    evid = Table[0, {nrn}];
    vtype = Switch[DefaultValueType,
      _List,	Take[Join[$ValueType/@DefaultValueType, Table[DBF$Native, {nrn - Length[DefaultValueType]}]], nrn],
      _,	With[{type = $ValueType[DefaultValueType]}, Table[type, {nrn}]]];
    count = Switch[DefaultCount,
      {__Real},	Take[Join[If[RealQ[#], #, 0]&/@DefaultCount, Table[0, {nrn - Length[DefaultCount]}]], nrn],
      _Real,	Table[DefaultCount, {nrn}],
      _,	Table[0, {nrn}]];

    FlushIO[Flush]);

  CreateChannel[pv_, opt:((_Rule|_RuleDelayed)...)] := (
    ClearChannel[];

    ReconfigureOptions[opt];

    rn = If[StringQ[pv] && StringLength[pv] > 0, pv, Null];
    nrn = -1;

    cs  = CS$NeverConn;
    st  = Undefined;
    sev = SEV$None;
    ts  = 0;
    v   = InitialValue;

    Clear[rnl, csl, sevl, tsl, vl, pos];

    chid = If[StringQ[rn], With[{id = Check[EPICS$CaOpen[rn], Null]},
      If[RealQ[id], rl[id] = {This, 0}]; id], Null];

    nact = RealQ[chid];

    (* Expand option for EPICS$CaAddEvent[] *)
    evid = 0;
    vtype = Switch[DefaultValueType,
      {__},	$ValueType[First[DefaultValueType]],
      _,	$ValueType[      DefaultValueType ]];
    count = Switch[DefaultCount,
      {__Real},	First[DefaultCount],
      _Real,	DefaultCount,
      _,	0];

    FlushIO[Flush]);

  ClearChannel[] := If[chid <=> Null,
    cs = If[nrn < 0,
      If[RealQ[chid], (EPICS$CaClearChannel[chid]; rl[chid] =.; CS$Closed), cs],
      MapThread[If[RealQ[#1], (EPICS$CaClearChannel[#1]; rl[#1] =.; CS$Closed), #2]&, {chid, cs}]];
    chid = Null];

  (* Decode ValueType into type number *)
  $ValueType[type_Real] := type;
  $ValueType[String]    := DBF$String;
  $ValueType[Real]      := DBF$Double;
  $ValueType[_]         := DBF$Native;

  (* Connection Test *)
  SetConStatCheckFlag[b_:True] := With[{prev = bcschk}, bcschk = b; prev];
  CheckConStat[] := Flatten[Map[If[#@nrn < 0,
    If[#@cs == CS$Conn, Null[], #@rn],
    If[MinMax[#@cs] == {CS$Conn, CS$Conn}, Null[],
      MapThread[If[#1 == CS$Conn, Null[], #2]&, {#@cs, #@rn}]]]&, il]];

  (* Last Event Lookup *)
  NameL[]       := rnl;
  ConStatL[]    := csl;
  StatusL[]     := stl;
  SeverityL[]   := sevl;
  TimeStampL$[] := tsl;
  TimeStampL[]  := tsl + EPICS$Epoch;
  ValueL[]      := vl;
  PositionL[]   := pos;

  ValueL[n_]      := vl[[n]];

  (* Current Value Lookup *)
  ConfiguredQ[]  := (chid <=> Null);
  Name[]         := rn;
  ValueType[]    := vtype;
  Count[]        := count;
  IOCName[]      := If[nrn < 0, EPICS$CaHostName[chid],     EPICS$CaHostName/@chid];
  FieldType[]    := If[nrn < 0, EPICS$CaFieldType[chid],    EPICS$CaFieldType/@chid];
  ElementCount[] := If[nrn < 0, EPICS$CaElementCount[chid], EPICS$CaElementCount/@chid];
  ConnectedQ[]   := If[nrn < 0, cs == CS$Conn, MapThread[Equal, {cs, CS$Conn}]];
  ConStat[]    := cs;
  Status[]     := st;
  Severity[]   := sev;
  TimeStamp$[] := ts;
  TimeStamp[]  := ts + EPICS$Epoch;
  Value[]      := v;
  ValueCopy[]  := If[nrn < 0, v, Duplicate[v]];

  (* Current Value Lookup in n-th element in monitor vector *)
  ConfiguredQ[n_Real]  := And[Length[chid] >= n, RealQ[chid[[n]]]];
  Name[n_Real]         := rn[[n]];
  ValueType[n_Real]    := vtype[[n]];
  Count[n_Real]        := count[[n]];
  IOCName[n_Real]      := EPICS$CaHostName[chid[[n]]];
  FieldType[n_Real]    := EPICS$CaFieldType[chid[[n]]];
  ElementCount[n_Real] := EPICS$CaElementCount[chid[[n]]];
  ConnectedQ[n_Real]   := (cs[[n]] == CS$Conn);
  ConStat[n_Real]    := cs[[n]];
  Status[n_Real]     := st[[n]];
  Severity[n_Real]   := sev[[n]];
  TimeStamp$[n_Real] := ts[[n]];
  TimeStamp[n_Real]  := ts[[n]] + EPICS$Epoch;
  Value[n__Real]     := v[[n]];
  ValueCopy[n__Real] := v[[n]];

  (* Modify runtime vtype/count field for EPICS$CaAddEvent[] *)
  SetValueType[v_] := With[{prev = vtype},
    vtype = If[nrn < 0,
      Switch[v,
	{__},	$ValueType[First[v]],
	_,	$ValueType[      v ]],
      Switch[v,
	_List,	Take[Join[$ValueType/@v, Table[DBF$Native, {nrn - Length[v]}]], nrn],
	_,	With[{type = $ValueType[v]}, Table[type, {nrn}]]]];
    prev];
  SetValueType[n_Real, v_] := With[{prev = vtype[[n]]}, vtype[[n]] = $ValueType[v]; prev];
  (ValueType[] = v_) ^:= SetValueType[v];
  (ValueType[n_Real] = v_) ^:= SetValueType[n, v];

  SetCount[v_]   := With[{prev = count},
    count = If[nrn < 0,
      Switch[v,
	{__Real},	First[v],
	_Real,		v,
	_,		count],
      Switch[v,
	{__Real},	Take[Join[If[RealQ[#], #, 0]&/@v, Table[0, {nrn - Length[v]}]], nrn],
	_Real,		Table[v, {nrn}],
	_,		count]];
    prev];
  SetCount[n_Real, v_Real] := With[{prev = count[[n]]}, count[[n]] = v; prev];
  (Count[] = v_) ^:= SetCount[v];
  (Count[n_Real] = v_) ^:= If[RealQ[v], SetCount[n, v], Null];

  (* Clear Value Buffer *)
  ClearValue[] := If[chid <=> Null,
    v = If[nrn < 0,
      InitialValue,
      Table[InitialValue, {nrn}]]];
  ClearValue[n_Real] := If[chid <=> Null, v[[n]] = InitialValue];

  (* Start & Stop monitor *)
  Start[] := (
    AutoStart = True;
    If[chid <=> Null,
      evid = If[nrn < 0,
	If[evid || cs <> CS$Conn, evid,
	  EPICS$CaAddEvent[chid, vtype, evflag, count]],
	MapThread[If[#2 || #1 <> CS$Conn, #2,
	  EPICS$CaAddEvent[#3, #4, evflag, #5]]&,
	  {cs, evid, chid, vtype, count}]];
      FlushIO[Flush]]);

  Stop[] := (
    AutoStart = False;
    If[chid <=> Null,
      evid = If[nrn < 0,
	(    If[evid, EPICS$CaClearEvent[evid]]; False),
	Map[(If[   #, EPICS$CaClearEvent[   #]]; False)&, evid]];
      FlushIO[Flush]]);

  Restart[] := If[chid <=> Null,
    evid = If[nrn < 0,
      (    If[evid, EPICS$CaClearEvent[evid]]; False),
      Map[(If[   #, EPICS$CaClearEvent[   #]]; False)&, evid]];
    FlushIO[]; Start[]];

  ClearRestart[] := If[chid <=> Null,
    evid = If[nrn < 0,
      (    If[evid, EPICS$CaClearEvent[evid]]; False),
      Map[(If[   #, EPICS$CaClearEvent[   #]]; False)&, evid]];
    ClearValue[];
    FlushIO[]; Start[]];

  Start[n_Real] := If[(evid[[n]] == 0) && (cs[[n]] == CS$Conn),
    evid[[n]] = EPICS$CaAddEvent[chid[[n]], vtype[[n]], evflag, count[[n]]]; FlushIO[Flush]];

  Stop[n_Real] := If[evid[[n]],
    EPICS$CaClearEvent[evid[[n]]]; evid[[n]] = False; FlushIO[Flush]];

  Restart[n_Real] := If[evid[[n]],
    EPICS$CaClearEvent[evid[[n]]]; evid[[n]] = False; FlushIO[]; Start[n]];
  ClearRestart[n_Real] := If[evid[[n]],
    EPICS$CaClearEvent[evid[[n]]]; evid[[n]] = False; FlushIO[]; ClearValue[n]; Start[n]];

  (* GetCB *)
  GetCB[] := Which[
    Not[ConfiguredQ[]],			Message[9, CA::unconfigured],
    nrn < 0,	(EPICS$CaGet[chid, count, vtype];		FlushIO[Flush]),
    True,	(ScanThread[EPICS$CaGet, {chid, count, vtype}];	FlushIO[Flush])];
  GetCB[n_Real] := Which[
    Not[ConfiguredQ[] && n =< nrn],	Message[9, CA::unconfigured],
    True,	(EPICS$CaGet[chid[[n]], count[[n]], vtype[[n]]]; FlushIO[Flush])];

  (* Put *)
  Put[v_List] := With[{n = Length[v]}, Which[
    Not[ConfiguredQ[]],			Message[9, CA::unconfigured],
    nrn < 0,	(EPICS$CaPut[chid, v];                          FlushIO[Flush]),
    n < nrn,	(ScanThread[EPICS$CaPut, {Take[chid, n], v}];   FlushIO[Flush]),
    n == nrn,	(ScanThread[EPICS$CaPut, {chid, v}];            FlushIO[Flush]),
    True,	(ScanThread[EPICS$CaPut, {chid, Take[v, nrn]}]; FlushIO[Flush])]];
  Put[v_] := Which[
    Not[ConfiguredQ[]],			Message[9, CA::unconfigured],
    True,	(EPICS$CaPut[chid,      v]; FlushIO[Flush])];
  Put[n_Real, v_] := Which[
    Not[ConfiguredQ[] && n =< nrn],	Message[9, CA::unconfigured],
    True,	(EPICS$CaPut[chid[[n]], v]; FlushIO[Flush])];

  (* PutCB *)
  PutCB[v_List] := With[{n = Length[v]}, Which[
    Not[ConfiguredQ[]],			Message[9, CA::unconfigured],
    nrn < 0,	(EPICS$CaPutCB[chid, v];                          FlushIO[Flush]),
    n < nrn,	(ScanThread[EPICS$CaPutCB, {Take[chid, n], v}];   FlushIO[Flush]),
    n == nrn,	(ScanThread[EPICS$CaPutCB, {chid, v}];            FlushIO[Flush]),
    True,	(ScanThread[EPICS$CaPutCB, {chid, Take[v, nrn]}]; FlushIO[Flush])]];
  PutCB[v_] := Which[
    Not[ConfiguredQ[]],			Message[9, CA::unconfigured],
    True,	(EPICS$CaPutCB[chid,      v]; FlushIO[Flush])];
  PutCB[n_Real, v_] := Which[
    Not[ConfiguredQ[] && n =< nrn],	Message[9, CA::unconfigured],
    True,	(EPICS$CaPutCB[chid[[n]], v]; FlushIO[Flush])];

  (* Backword compatibility method[obsolete] *)
  CheckAndPut[v_] := If[cs <=> CS$Conn, Print["Put Error: ",rn," is not connected !"],
    EPICS$CaPut[chid, v]];
  CheckAndPutL[n_Real, v_] := If[cs[[n]] <=> CS$Conn, Print["Put Error: ", rn[[n]], " is not connected !"],
    EPICS$CaPut[chid[[n]], v]];

  (* Check & Wait Value[] filling *)
  CountUndef[] := Which[
    chid === Null,	0,					(* No configured PV or ClearChannel-ed *)
    nrn < 0,		(Undefined === v),			(* Configured PV *)
    nact == nrn,	Literal[Count][v, Undefined, {1}],	(* Full configured PV vector *)
    True,		Plus@@MapThread[And[RealQ[#1], Undefined === #2]&, {chid, v}]];

  (* CountUndef for given instance list *)
  CountUndef[il_List] := Plus@@Map[#@CountUndef[]&, il];

  With[{dT = 100e-3},	(* Step time to wait ValueCB events *)

    BreakWaitValue[] := If[TkinterQ[],
      (BreakWaitValue[] := If[waitflg, (waitflg = False; TkReturn[])]; BreakWaitValue[]),
      If[waitflg, waitflg = False]];

    WaitValue[timeout_Real:0] := If[TkinterQ[],
      (If[This === CaMonitor,
	WaitValue[timeout$_Real:0] := WaitValueByTkSense[timeout$, Instances[]],
	WaitValue[timeout$_Real:0] := WaitValueByTkSense[timeout$]]; WaitValue[timeout]),
      If[This === CaMonitor, WaitValueBySleep[timeout, Instances[]], WaitValueBySleep[timeout]]];

    (* WaitValue for Self *)
    WaitValueByTkSense[timeout_Real:0] := With[{limit = FromDate[Tick->0] + timeout},
      waitflg = True;		(* Enter WaitValue section *)
      With[{success = (If[timeout > 0,
	While[waitflg && CountUndef[] > 0, With[{t = FromDate[Tick->0]},
	  If[limit > t, TkSense[Min[dT, limit - t]], waitflg = False]]],
	While[waitflg && CountUndef[] > 0, TkSense[dT]]]; waitflg)},
	waitflg = False;	(* Exit WaitValue section *)
	success]];

    WaitValueBySleep[timeout_Real:0] := With[{limit = FromDate[Tick->0] + timeout},
      waitflg = True;		(* Enter WaitValue section *)
      With[{success = (If[timeout > 0,
	While[waitflg && CountUndef[] > 0, With[{t = FromDate[Tick->0]},
	  If[limit > t, EPICS$CaPendEvent[Min[dT, limit - t]], waitflg = False]]],
	While[waitflg && CountUndef[] > 0, EPICS$CaPendEvent[dT]]]; waitflg)},
	waitflg = False;	(* Exit WaitValue section *)
	success]];

    (* WaitValue for given instance list *)
    WaitValue[timeout_Real:0, il_List] := If[TkinterQ[],
      (WaitValue[timeout$_Real:0, il$_List] := WaitValueByTkSense[timeout$, il$]; WaitValue[timeout, il]),
      WaitValueBySleep[timeout, il]];

    WaitValueByTkSense[timeout_Real:0, il_List] := With[{limit = FromDate[Tick->0] + timeout},
      waitflg = True;		(* Enter WaitValue section *)
      With[{success = (If[timeout > 0,
	While[waitflg && CountUndef[il] > 0, With[{t = FromDate[Tick->0]},
	  If[limit > t, TkSense[Min[dT, limit - t]], waitflg = False]]],
	While[waitflg && CountUndef[il] > 0, TkSense[dT]]]; waitflg)},
	waitflg = False;	(* Exit WaitValue section *)
	success]];

    WaitValueBySleep[timeout_Real:0, il_List] := With[{limit = FromDate[Tick->0] + timeout},
      waitflg = True;		(* Enter WaitValue section *)
      With[{success = (If[timeout > 0,
	While[waitflg && CountUndef[il] > 0, With[{t = FromDate[Tick->0]},
	  If[limit > t, EPICS$CaPendEvent[Min[dT, limit - t]], waitflg = False]]],
	While[waitflg && CountUndef[il] > 0, EPICS$CaPendEvent[dT]]]; waitflg)},
	waitflg = False;	(* Exit WaitValue section *)
	success]];

    ];

  (* House keeping ca_pend_event loop for eca_fd_callback stacking case *)
  DoPendEvent[force_Real:False] := With[{t = FromDate[Tick->0]}, Which[
    $CaInPend,				After[10e-3, DoPendEvent[force]],

    t - lastPendEvent > PendEventDelay,	(
      EPICS$CaPendEvent[]; lastPendEvent = t;
      After[1.25 * PendEventDelay, DoPendEvent[]]),

    force,				(
      After[1.25 * PendEventDelay, DoPendEvent[]])]];

  CheckDoPendEvent[] := If[afterPendEvent === Null && FromDate[Tick->0] - lastPendEvent > PendEventDelay,
    afterPendEvent = AfterIdle[afterPendEvent = Null; DoPendEvent[True]]];

  ];	(* End of CaMonitor class *)

! Register CaMonitor@DoPendEvent[True] to Tkinter hook
PkgHookAdd["Tkinter", CaMonitor@DoPendEvent[True]];

!! EPICS$ConStatCB[chid_Real, status_Real] callback for EPICS CA connection handler: eca_constatcb,
!! however, this function is never invoked by CaMonitor backend part(src/tfEPICSCA.c)
!! because eca_constatcb calls Fortran version of this callback: CaMonitor_ConStatCB @ src/epicsca_callback.f
EPICS$ConStatCB[chid_Real, status_Real] := With[{a = CaMonitor@rl[chid][[1]], n = CaMonitor@rl[chid][[2]]},
  If[n < 1,
    ( a@cs = status),
    ( a@pos = n;
      a@cs[[n]] = status;
      a@rnl = a@rn[[n]];
      a@csl = status)];
  If[Head[a@cscomm] === Hold, ReleaseHold[a@cscomm]];
  If[a@AutoStart && (status === CaMonitor@CS$Conn),
    If[n >= 1, a@Start[n], a@Start[]]]];

!! EPICS$PutCB[chid_Real] callback for EPICS CA put event handler: eca_putcb,
!! however, this function is never invoked by CaMonitor backend part(src/tfEPICSCA.c)
!! because eca_putcb calls Fortran version of this callback: CaMonitor_PutCB @ src/epicsca_callback.f
EPICS$PutCB[chid_Real] := With[{a = CaMonitor@rl[chid][[1]], n = CaMonitor@rl[chid][[2]]},
  If[n < 1,
    ( Null ),
    ( a@pos = n;
      a@rnl = a@rn[[n]])];
  If[Head[a@putcomm] === Hold, ReleaseHold[a@putcomm]]];

!! EPICS$ValueCB[chid_Real, value_, {status-of-value_Real, alarm-severity_Real, time-stamp_Real}] callback
!! for EPICS CA connection handler: eca_valuecb,
!! however, this function is never invoked by CaMonitor backend part(src/tfEPICSCA.c)
!! because eca_valuecb calls Fortran version of this callback: CaMonitor_ValueCB @ src/epicsca_callback.f
EPICS$ValueCB[chid_Real, value_, {status_Real, severity_Real, stamp_Real}] := With[{
  a = CaMonitor@rl[chid][[1]], n = CaMonitor@rl[chid[[2]]]},
  If[n < 1,
    ( a@v = value;
      a@st = status;
      a@sev = severity;
      a@ts = stamp),
    ( a@pos = n;
      a@v[[n]] = value;
      a@st[[n]] = status;
      a@sev[[n]] = severity;
      a@ts[[n]] = stamp;
      a@rnl = a@rn[[n]];
      a@vl = value;
      a@stl = status;
      a@sevl = severity;
      a@tsl = stamp)];
  If[Head[a@valcomm] === Hold, ReleaseHold[a@valcomm]]];

! End of File
